from __future__ import annotations

import numpy as np
from scipy.optimize import minimize
from torch.nn import Softmax

from garrus.const import EPS
from garrus.core import BaseCalibration
from garrus.metrics import NLL


class Platt(BaseCalibration):
    def __init__(self):
        super().__init__()

    def _fit(self, confidences: np.ndarray, accuracies: np.ndarray, **kwargs) -> Platt:
        X = self._scipy_transform_data(confidences)

        weights_num = 2
        x0 = np.ones(weights_num)
        x0[0] = EPS

        result = minimize(fun=self._loss_func, x0=x0, args=(X, accuracies))

        self._weights = result.x
        return self

    def _transform(self, confidences: np.ndarray, **kwargs) -> np.ndarray:
        X = self._scipy_transform_data(confidences)
        logits = self._update_weights(X, self._weights)
        calibrated = Softmax(dim=1)(logits).numpy()
        return calibrated

    def _loss_func(self, weights: np.ndarray, X: np.ndarray, y: np.ndarray):
        logits = self._update_weights(X, weights)
        return NLL().compute(logits, y)



from typing import Union

import numpy as np
import plotly.graph_objects as go

from garrus.core import BaseVisualization


class ConfidenceHistogram(BaseVisualization):
    def _plot(self, confidences: np.ndarray, accuracies: np.ndarray, **kwargs: Union[int, float]) -> None:
        fig = go.Figure()

        fig.add_trace(
            go.Histogram(
                x=confidences,
                xbins=dict(start=0, end=1.0, size=1 / self.n_bins),
                name="Amount of samples"
            )
        )

        fig.add_vline(
            x=confidences.mean(),
            line_width=3, line_dash="dash", line_color="green",
            name="Avg. confidence"
        )
        fig.add_trace(
            go.Scatter(
                x=[confidences.mean()], y=[0], mode="markers", name="Avg. confidence",
                marker={"color": "green", "size": 10}
            )
        )

        fig.add_vline(
            x=accuracies.mean(),
            line_width=3, line_dash="dash", line_color="red",
            name="Avg. accuracy"
        )
        fig.add_trace(
            go.Scatter(
                x=[accuracies.mean()], y=[0], mode="markers", name="Avg. accuracy",
                marker={"color": "red", "size": 10}
            )
        )

        fig.update_yaxes(rangemode="tozero")

        fig.update_layout(
            title="Confidence Histogram",
            xaxis_title="Confidence",
            yaxis_title="Samples",
            showlegend=True,
            height=self.plot_h, width=self.plot_w,
        )

        fig.show()



from .ece import ECE
from .nll import NLL
from .aurc import AURC
from .fpr_n_tpr import FPRnTPR
from .eaurc import EAURC
from .brier import Brier
from .aupre import AUPRE

__all__ = ["ECE", "NLL", "AURC", "AUPRE", "EAURC", "Brier", "FPRnTPR"]



import numpy as np

from garrus.core import BaseMetric
from garrus.metrics.aurc import coverage_risk


class EAURC(BaseMetric):
    """
    Excess Area Under Risk-Coverage.

    $$ E-AURC = AURC (\kappa, f \mid V_{n}) - AURC (\kappa^{*}, f \mid V_{n}) $$
    """

    def _compute(self, confidences: np.ndarray, accuracies: np.ndarray, **kwargs) -> float:
        sort_values = sorted(zip(confidences, accuracies), key=lambda x: x[0], reverse=True)
        sort_conf, sort_acc = zip(*sort_values)
        risk_list, coverage_list = coverage_risk(sort_conf, sort_acc)

        r = risk_list[-1]
        risk_coverage_curve_area = 0
        optimal_risk_area = r + (1 - r) * np.log(1 - r)
        for risk_value in risk_list:
            risk_coverage_curve_area += risk_value * (1 / len(risk_list))

        eaurc = risk_coverage_curve_area - optimal_risk_area
        return float(eaurc)



import numpy as np

from garrus.core import BaseMetric


class ECE(BaseMetric):
    """
    Expected Calibration Error.

    $$ ECE=\frac{1}{|\{B\}| m} \sum_{B \in\{B\}} \mid \sum_{x \in B} b_{k}(x)-\sum_{x \in B} I[y(x)=k] $$,
    for fixed class k.
    """

    def _compute(self, confidences: np.ndarray, accuracies: np.ndarray, **kwargs) -> float:
        bins = np.linspace(0, 1, self.n_bins + 1)  # noqa
        bin_lowers = bins[:-1]
        bin_uppers = bins[1:]

        ece = np.zeros(1)

        for bin_lower, bin_upper in zip(bin_lowers, bin_uppers):
            in_bin = (confidences > bin_lower.item()) * (confidences <= bin_upper.item())
            prop_in_bin = in_bin.astype(float).mean()

            if prop_in_bin.item() > 0.0:
                accuracy_in_bin = accuracies[in_bin].astype(float).mean()
                avg_confidence_in_bin = confidences[in_bin].mean()

                ece += np.abs(avg_confidence_in_bin - accuracy_in_bin) * prop_in_bin

        return float(ece.item())



import numpy as np
from sklearn import metrics

from garrus.core import BaseMetric


class FPRnTPR(BaseMetric):
    """False positive rate at n% true positive rate."""

    def _compute(self, confidences: np.ndarray, accuracies: np.ndarray, **kwargs: float) -> float:
        """
        Keyword Args:
            n (float [0, 1]): true positive rate %.
        """
        fpr, tpr, thresholds = metrics.roc_curve(accuracies, confidences)
        idx_tpr = np.argmin(np.abs(tpr - kwargs["n"]))  # noqa
        fpr_n_tpr = fpr[idx_tpr]

        return float(fpr_n_tpr)



from typing import Union, Optional
from abc import ABC, abstractmethod

import numpy as np


class BaseMetric(ABC):
    def __init__(self, n_bins: Optional[int] = None):
        self.n_bins = n_bins

    @abstractmethod
    def _compute(self, confidences: np.ndarray, accuracies: np.ndarray, **kwargs) -> float:
        """
        Private metric computation method.

        Args:
             confidences (np.ndarray): array of confidence levels for each sample;
             accuracies (np.ndarray): array of 0/1 labels of correctness for each sample;
             kwargs (Any): methods parameters.
        Returns:
            float: calculated metric.
        """

    def compute(self, confidences: np.ndarray, accuracies: np.ndarray, **kwargs: Union[int, float]) -> float:
        """
        Main metric computation method.

        Args:
             confidences (np.ndarray): array of confidence levels for each sample;
             accuracies (np.ndarray): array of 0/1 labels of correctness for each sample;
             kwargs (Any): methods parameters.
        Returns:
            float: calculated metric.
        """
        assert confidences.shape[0] == accuracies.shape[0], "Number of conf and acc samples is not equal."
        assert accuracies.size > 0, "No samples provided."

        metric = self._compute(confidences, accuracies, **kwargs)
        return metric



import numpy as np
from scipy.special import expit

from garrus.const import EPS
from garrus.core import BaseMetric


class NLL(BaseMetric):
    """
    Negative Log-Likelihood.

    $$ NLL = -\frac{1}{m} \sum_{j=1}^{m} y_{j} \log b_{j} $$
    """

    def _compute(self, confidences: np.ndarray, accuracies: np.ndarray, **kwargs) -> float:
        if confidences.ndim == 2:
            confidences = confidences[:, -1]

        confidences = expit(confidences)
        confidences = np.clip(confidences, EPS, 1. - EPS)  # noqa

        if confidences.ndim == 2:
            confidences = np.array(confidences[:, -1])

        cross_entropy = np.multiply(accuracies, np.log(confidences)) + \
                        np.multiply(1. - accuracies, np.log(1 - confidences))

        log_likelihood = np.sum(cross_entropy)  # noqa
        nll = -log_likelihood / accuracies.shape[0]
        return nll



import numpy as np
from sklearn import metrics

from garrus.core import BaseMetric


class AUPRE(BaseMetric):
    """Area under the precision-recall curve using errors as the positive class."""

    def _compute(self, confidences: np.ndarray, accuracies: np.ndarray, **kwargs) -> float:
        aupr_err = metrics.average_precision_score(-1 * accuracies + 1, -1 * confidences)

        return float(aupr_err)



from typing import Union, Tuple, List

import numpy as np
import plotly.graph_objects as go

from garrus.core import BaseVisualization


class ReliabilityDiagram(BaseVisualization):
    def __calc_statistics(
            self, confidences: np.ndarray, accuracies: np.ndarray
    ) -> Tuple[List[float], List[float], List[float]]:
        mean_conf_by_bin: List[float] = []
        mean_acc_by_bin: List[float] = []
        samples_pct: List[float] = []

        bin_idxs = np.digitize(confidences, np.histogram_bin_edges(confidences, bins=self.n_bins))

        for bin_idx in range(1, 11):
            mean_conf_by_bin.append(confidences[bin_idxs == bin_idx].mean() \
                                        if any(confidences[bin_idxs == bin_idx]) else 0)
            mean_acc_by_bin.append(accuracies[bin_idxs == bin_idx].mean() \
                                       if any(accuracies[bin_idxs == bin_idx]) else 0)
            samples_pct.append((bin_idxs == bin_idx).sum() / bin_idxs.shape[0])

        return mean_conf_by_bin, mean_acc_by_bin, samples_pct

    def _plot(self, confidences: np.ndarray, accuracies: np.ndarray, **kwargs: Union[int, float]) -> None:
        mean_conf_by_bin, mean_acc_by_bin, samples_pct = self.__calc_statistics(confidences, accuracies)

        bars = [x / self.n_bins for x in range(self.n_bins + 1)]

        fig = go.Figure()

        # confidence --------------
        fig.add_trace(
            go.Bar(
                y=mean_conf_by_bin, x=bars, name="Mean confidence",
                marker={"color": "red", "opacity": 0.6}
            ),
        )

        # accuracy ----------------
        fig.add_trace(
            go.Bar(
                y=mean_acc_by_bin, x=bars, name="Mean accuracy",
                marker={"color": "blue", "opacity": 0.6}
            ),
        )
        fig.add_trace(
            go.Scatter(
                y=mean_acc_by_bin, x=bars, name="Mean confidence dot",
                marker={"color": "blue", "size": 9}, mode="markers+lines", showlegend=False
            ),
        )

        # % of samples ------------
        fig.add_trace(
            go.Bar(
                y=samples_pct, x=bars, name="% of samples", width=0.025,
                marker={"color": "black", "opacity": 1.0}
            ),
        )

        # ideal line --------------
        fig.add_shape(
            type="line",
            x0=-0.05, y0=0, x1=0.95, y1=1,
            line=dict(
                color="black",
                width=2,
                dash="dot",
            )
        )

        fig.update_yaxes(rangemode="tozero")

        fig.update_layout(
            title="Reliability Plot",
            xaxis_title="Confidence",
            yaxis_title="Accuracy",
            showlegend=True,
            barmode="overlay",
            bargap=0,
            height=self.plot_h, width=self.plot_w,
        )

        fig.show()



from __future__ import annotations
from abc import abstractmethod

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from scipy.special import logit

from garrus.const import EPS


class BaseCalibration(BaseEstimator, TransformerMixin):
    def __init__(self):
        super().__init__()

        self._weights = None
        self.__fitted = False
        self.__accuracies = np.array([])

    @abstractmethod
    def _fit(self, confidences: np.ndarray, accuracies: np.ndarray, **kwargs) -> BaseCalibration:
        """
        Private calibration fit method.

        Args:
             confidences (np.ndarray): array of confidence levels for each sample;
             accuracies (np.ndarray): array of 0/1 labels of correctness for each sample;
             kwargs (Any): methods parameters.
        """

    def fit(self, confidences: np.ndarray, accuracies: np.ndarray, **kwargs) -> BaseCalibration:
        """
        Main calibration fit method.

        Args:
             confidences (np.ndarray): array of confidence levels for each sample;
             accuracies (np.ndarray): array of 0/1 labels of correctness for each sample;
             kwargs (Any): methods parameters.
        """
        assert confidences.shape[0] == accuracies.shape[0], "Number of conf and acc samples is not equal."
        assert accuracies.size > 0, "No samples provided."
        self.__accuracies = accuracies

        return self._fit(confidences, accuracies, **kwargs)

    @abstractmethod
    def _transform(self, confidences: np.ndarray, **kwargs) -> np.ndarray:
        """
        Private calibration transform method.

        Args:
             confidences (np.ndarray): array of confidence levels for each sample;
             kwargs (Any): methods parameters.
        """

    def transform(self, confidences: np.ndarray, **kwargs) -> np.ndarray:
        """
        Main calibration transform method.

        Args:
             confidences (np.ndarray): array of confidence levels for each sample;
             kwargs (Any): methods parameters.
        """
        assert self.__fitted, "Should call fit() before making transform"
        assert confidences.shape[0] == self.__accuracies.shape[0], "Number of conf and acc samples is not equal."

        return self._transform(confidences, **kwargs)

    def _scipy_transform_data(self, confidences: np.ndarray) -> np.ndarray:
        """
        Transforms data input for calibrations methods utilize scipy optimizations

        Args:
             confidences (np.ndarray): array of confidence levels for each sample;
        Returns:
            np.ndarray: transformed data for scipy input.
        """
        if confidences.ndim == 1:
            confidences = confidences.reshape(-1, 1)
        confidences = np.clip(confidences, EPS, 1. - EPS)  # noqa
        X = logit(confidences)
        return X

    def _update_weights(self, X: np.ndarray, weights: np.ndarray):
        """
        Performs weights update by given data input and weights.
        """
        bias = weights[0]
        weights = np.array(weights[1:]).reshape(-1, 1)
        logits = np.matmul(X, weights) + bias
        return logits



import numpy as np

from garrus.core import BaseMetric


class Brier(BaseMetric):
    """
    Brier score.

    $$ Brier = -\frac{1}{m} \sum_{j=1}^{m} (y_{j}-b_{j})^{2}) $$
    """

    def _compute(self, confidences: np.ndarray, accuracies: np.ndarray, **kwargs) -> float:
        brier_score = np.mean(np.sum((confidences - accuracies) ** 2, axis=1))  # noqa
        return float(brier_score)



from typing import Union
from abc import ABC, abstractmethod

import numpy as np


class BaseVisualization(ABC):
    def __init__(self, n_bins: int, plot_h: int = 650, plot_w: int = 800):
        self.n_bins = n_bins

        self.plot_h = plot_h
        self.plot_w = plot_w

    @abstractmethod
    def _plot(self, confidences: np.ndarray, accuracies: np.ndarray, **kwargs: Union[int, float]) -> None:
        """
        Private visualization plotting method.

        Args:
             confidences (np.ndarray): array of confidence levels for each sample;
             accuracies (np.ndarray): array of 0/1 labels of correctness for each sample;
             kwargs (Any): methods parameters.
        """

    def plot(self, confidences: np.ndarray, accuracies: np.ndarray, **kwargs: Union[int, float]) -> None:
        """
        Main visualization plotting method.

        Args:
             confidences (np.ndarray): array of confidence levels for each sample;
             accuracies (np.ndarray): array of 0/1 labels of correctness for each sample;
             kwargs (Any): methods parameters.
        """
        assert confidences.shape[0] == accuracies.shape[0], "Number of conf and acc samples is not equal."
        assert accuracies.size > 0, "No samples provided."

        self._plot(confidences, accuracies, **kwargs)



#!/usr/bin/env python
# flake8: noqa
# -*- coding: utf-8 -*-

# Note: To use the "upload" functionality of this file, you must:
#   $ pip install twine

import io
import os
import sys
import subprocess
from shutil import rmtree

from setuptools import Command, find_packages, setup

# Package meta-data.
NAME = "garrus"
DESCRIPTION = "Garrus. Python framework for better confidence estimate of deep neural networks."
URL = "https://github.com/sleep3r/garrus"
EMAIL = "sleep3r@icloud.com"
AUTHOR = "Alexander Kalashnikov"
REQUIRES_PYTHON = ">=3.7.0"

PROJECT_ROOT = os.path.abspath(os.path.dirname(__file__))


def load_requirements(filename):
    """Docs? Contribution is welcome."""
    with open(os.path.join(PROJECT_ROOT, filename), "r") as f:
        return f.read().splitlines()


def load_readme():
    """Docs? Contribution is welcome."""
    readme_path = os.path.join(PROJECT_ROOT, "README.md")
    with io.open(readme_path, encoding="utf-8") as f:
        return f"\n{f.read()}"


def load_version():
    """Docs? Contribution is welcome."""
    context = {}
    with open(os.path.join(PROJECT_ROOT, "garrus", "__init__.py")) as f:
        exec(f.read(), context)
    return context["__version__"]


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def run(self):
        """Docs? Contribution is welcome."""
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(PROJECT_ROOT, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        subprocess.call("{0} setup.py sdist bdist_wheel --universal".format(sys.executable), shell=False)

        self.status("Uploading the package to PyPI via Twine…")
        subprocess.call("twine upload dist/*", shell=False)

        self.status("Pushing git tags…")
        subprocess.call("git tag v{0}".format(load_version()), shell=False)
        subprocess.call("git push --tags", shell=False)

        sys.exit()


setup(
    name=NAME,
    version=load_version(),
    description=DESCRIPTION,
    long_description=load_readme(),
    long_description_content_type="text/markdown",
    keywords=[
        "Machine Learning",
        "Distributed Computing",
        "Deep Learning",
        "Reinforcement Learning",
        "Computer Vision",
        "Natural Language Processing",
        "Recommendation Systems",
        "Information Retrieval",
        "PyTorch",
        "Confidence calibration",
        "Confidence ranking"
    ],
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    download_url=URL,
    project_urls={
        "Bug Tracker": URL + "/issues",
        "Documentation": URL + "/wiki",
        "Source Code": URL ,
    },
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    license="Apache License 2.0",
    classifiers=[
        "Environment :: Console",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License",
        # Audience
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        # Topics
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis",
        # Programming
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    # $ setup.py publish support.
    cmdclass={"upload": UploadCommand},
)



import numpy as np

from garrus.core import BaseMetric


def coverage_risk(confidences, accuracies):
    risk_list = []
    coverage_list = []
    risk = 0
    for i in range(len(confidences)):
        coverage = (i + 1) / len(confidences)
        coverage_list.append(coverage)

        if accuracies[i] == 0:
            risk += 1

        risk_list.append(risk / (i + 1))
    return risk_list, coverage_list


class AURC(BaseMetric):
    """
    Area Under Risk-Coverage.

    $$ AURC (\kappa, f \mid V_{n}) = \frac{1}{n} \sum_{\theta \in \Theta} \hat{r} (f, g_{\theta} \mid V_{n}) $$
    """

    def _compute(self, confidences: np.ndarray, accuracies: np.ndarray, **kwargs) -> float:
        sort_values = sorted(zip(confidences, accuracies), key=lambda x: x[0], reverse=True)
        sort_conf, sort_acc = zip(*sort_values)
        risk_list, coverage_list = coverage_risk(sort_conf, sort_acc)

        risk_coverage_curve_area = 0
        for risk_value in risk_list:
            risk_coverage_curve_area += risk_value * (1 / len(risk_list))

        aurc = risk_coverage_curve_area
        return float(aurc)
